@@ -63,6 +63,7 @@ static int _log_suppress = 0;
 static int _control_fd = -1;
 static int _version_checked = 0;
 static int _version_ok = 1;
+static unsigned _ioctl_buffer_double_factor = 0;
 
 /*
  * Support both old and new major numbers to ease the transition.
@@ -610,8 +611,10 @@ static int _dm_task_run_v1(struct dm_task *dmt)
 	if (dmt->type == DM_DEVICE_TABLE)
 		dmi->flags |= DM_STATUS_TABLE_FLAG;
 
-	log_debug("dm %s %s %s %s", _cmd_data_v1[dmt->type].name, dmi->name,
-		  dmi->uuid, dmt->newname ? dmt->newname : "");
+	log_debug("dm %s %s %s%s%s [%u]", _cmd_data_v1[dmt->type].name,
+		  dmi->name, dmi->uuid, dmt->newname ? " " : "",
+		  dmt->newname ? dmt->newname : "",
+		  dmi->data_size);
 	if (dmt->type == DM_DEVICE_LIST) {
 		if (!_dm_names_v1(dmi))
 			goto bad;
@@ -1026,7 +1029,7 @@ static void *_add_target(struct target *t, void *out, void *end)
 	return out;
 }
 
-static struct dm_ioctl *_flatten(struct dm_task *dmt)
+static struct dm_ioctl *_flatten(struct dm_task *dmt, unsigned repeat_count)
 {
 	const size_t min_size = 16 * 1024;
 	const int (*version)[3];
@@ -1077,6 +1080,10 @@ static struct dm_ioctl *_flatten(struct dm_task *dmt)
 	if (len < min_size)
 		len = min_size;
 
+	/* Increase buffer size if repeating because buffer was too small */
+	while (repeat_count--)
+		len *= 2;
+
 	if (!(dmi = malloc(len)))
 		return NULL;
 
@@ -1276,10 +1283,59 @@ static int _create_and_load_v4(struct dm_task *dmt)
 	return r;
 }
 
+static struct dm_ioctl *_do_dm_ioctl(struct dm_task *dmt, unsigned command,
+				     unsigned repeat_count)
+{
+	struct dm_ioctl *dmi;
+
+	dmi = _flatten(dmt, repeat_count);
+	if (!dmi) {
+		log_error("Couldn't create ioctl argument");
+		return NULL;
+	}
+
+	if (dmt->type == DM_DEVICE_TABLE)
+		dmi->flags |= DM_STATUS_TABLE_FLAG;
+
+	dmi->flags |= DM_EXISTS_FLAG;	/* FIXME */
+
+	if (dmt->no_open_count)
+		dmi->flags |= DM_SKIP_BDGET_FLAG;
+
+	log_debug("dm %s %s %s%s%s %c %.0llu %s [%u]",
+		  _cmd_data_v4[dmt->type].name,
+		  dmi->name, dmi->uuid, dmt->newname ? " " : "",
+		  dmt->newname ? dmt->newname : "",
+		  dmt->no_open_count ? 'N' : 'O',
+		  dmt->sector, dmt->message ? dmt->message : "",
+		  dmi->data_size);
+#ifdef DM_IOCTLS
+	if (ioctl(_control_fd, command, dmi) < 0) {
+		if (errno == ENXIO && ((dmt->type == DM_DEVICE_INFO) ||
+				       (dmt->type == DM_DEVICE_MKNODES)))
+			dmi->flags &= ~DM_EXISTS_FLAG;	/* FIXME */
+		else {
+			if (_log_suppress)
+				log_verbose("device-mapper ioctl "
+					    "cmd %d failed: %s",
+					    _IOC_NR(command), strerror(errno));
+			else
+				log_error("device-mapper ioctl "
+					  "cmd %d failed: %s",
+					  _IOC_NR(command), strerror(errno));
+			free(dmi);
+			return NULL;
+		}
+	}
+#else /* Userspace alternative for testing */
+#endif
+	return dmi;
+}
+
 int dm_task_run(struct dm_task *dmt)
 {
-	struct dm_ioctl *dmi = NULL;
-	unsigned int command;
+	struct dm_ioctl *dmi;
+	unsigned command;
 
 #ifdef DM_COMPAT
 	if (_dm_version == 1)
@@ -1290,7 +1346,7 @@ int dm_task_run(struct dm_task *dmt)
 	    (sizeof(_cmd_data_v4) / sizeof(*_cmd_data_v4))) {
 		log_error("Internal error: unknown device-mapper task %d",
 			  dmt->type);
-		goto bad;
+		return 0;
 	}
 
 	command = _cmd_data_v4[dmt->type].cmd;
@@ -1306,43 +1362,25 @@ int dm_task_run(struct dm_task *dmt)
 	if (!_open_control())
 		return 0;
 
-	dmi = _flatten(dmt);
-	if (!dmi) {
-		log_error("Couldn't create ioctl argument");
+repeat_ioctl:
+	if (!(dmi = _do_dm_ioctl(dmt, command, _ioctl_buffer_double_factor)))
 		return 0;
-	}
 
-	if (dmt->type == DM_DEVICE_TABLE)
-		dmi->flags |= DM_STATUS_TABLE_FLAG;
-
-	dmi->flags |= DM_EXISTS_FLAG;	/* FIXME */
-
-	if (dmt->no_open_count)
-		dmi->flags |= DM_SKIP_BDGET_FLAG;
-
-	log_debug("dm %s %s %s %s%c %.0llu %s", _cmd_data_v4[dmt->type].name,
-		  dmi->name, dmi->uuid, dmt->newname ? dmt->newname : "",
-		  dmt->no_open_count ? 'N' : 'O',
-		  dmt->sector, dmt->message ? dmt->message : "");
-#ifdef DM_IOCTLS
-	if (ioctl(_control_fd, command, dmi) < 0) {
-		if (errno == ENXIO && ((dmt->type == DM_DEVICE_INFO) ||
-				       (dmt->type == DM_DEVICE_MKNODES)))
-			dmi->flags &= ~DM_EXISTS_FLAG;	/* FIXME */
-		else {
-			if (_log_suppress)
-				log_verbose("device-mapper ioctl "
-					    "cmd %d failed: %s",
-					    _IOC_NR(command), strerror(errno));
-			else
-				log_error("device-mapper ioctl "
-					  "cmd %d failed: %s",
-					  _IOC_NR(command), strerror(errno));
-			goto bad;
+	if (dmi->flags & DM_BUFFER_FULL_FLAG) {
+		switch (dmt->type) {
+		case DM_DEVICE_LIST_VERSIONS:
+		case DM_DEVICE_LIST:
+		case DM_DEVICE_DEPS:
+		case DM_DEVICE_STATUS:
+		case DM_DEVICE_TABLE:
+		case DM_DEVICE_WAITEVENT:
+			_ioctl_buffer_double_factor++;
+			free(dmi);
+			goto repeat_ioctl;
+		default:
+			log_error("Warning: libdevmapper buffer too small for data");
 		}
 	}
-#else /* Userspace alternative for testing */
-#endif
 
 	switch (dmt->type) {
 	case DM_DEVICE_CREATE:
